@@ -94,7 +94,7 @@ def unit_separator(math_string, supported_values, unit_identifier=' '):
     # clean-up symbols from scientific notation and remove excess math symbols at the end
     unit_expression = split_list.copy()
     i0 = 0
-    while i0 < len(unit_expression):
+    while unit_expression and i0 < len(unit_expression):
         i0_val = unit_expression[i0]  # define current value
         ip1 = i0 + 1  # define right value index
         try:  # try to define real right value, otherwise set to None
@@ -407,6 +407,13 @@ def unit_solver(unit_expression, unit_list, fixed_unit_list):
     i0 = 0
     while i0 < max_iteration:
         i0_val = unit_expression[i0]  # define current value
+        ip1_val = ip2_val = None  # define initial right values
+        try:
+            ip1_val = unit_expression[i0 + 1]
+            ip2_val = unit_expression[i0 + 2]
+        except IndexError:
+            pass
+
         if i0_val in unit_list:
 
             # at the same time expand expression around multiplication, and place it in parentheses if not already
@@ -414,15 +421,15 @@ def unit_solver(unit_expression, unit_list, fixed_unit_list):
                 im1_val = unit_expression[i0 - 1]
             except IndexError:
                 im1_val = None
-            try:
-                ip1_val = unit_expression[i0 + 1]
-            except IndexError:
-                ip1_val = None
 
             if im1_val == '(' and ip1_val == ')':
                 base_units_expression += nsu.split([fixed_unit_list[unit_list.index(i0_val)]][0], '*')
             else:
                 base_units_expression += ['('] + nsu.split([fixed_unit_list[unit_list.index(i0_val)]][0], '*') + [')']
+        elif (i0_val, ip1_val, ip2_val) == ('/', '(', ')'):
+            i0 += 2
+        elif (i0_val, ip1_val) == ('(', ')'):
+            i0 += 1
         else:
             base_units_expression += [i0_val]
         i0 += 1
@@ -717,13 +724,17 @@ def unit_parser(math_string, unit_identifier=' ', result='math', **kwargs):
             natively supported by this script.
         abb_unit : bool
             Specify whether the script should attempt to condense the resulting unit into a derived SI unit.
+        sf : int
+            Sets the significant figures of the result. Uses mpmath to do so. If set to None (which is default) no
+            attempt will be made to set significant figures. Note that this only affects the result printed in the
+            console, it does not whatsoever alter the return value.
 
     Returns
         The computed result as set by the result parameter.
     """
 
     # check and fix kwargs
-    parser_true_string, parser_print, abb_unit = math_string, 'num', True
+    parser_true_string, parser_print, abb_unit, sf = math_string, 'num', True, None
     if 'true_string' in kwargs.keys():
         parser_true_string = kwargs.get('true_string')
     if 'cprint' in kwargs.keys():
@@ -738,6 +749,8 @@ def unit_parser(math_string, unit_identifier=' ', result='math', **kwargs):
             for j in [unit_identifier + i for i in prefix_scalars]:
                 supported_units.append(j + i)
         supported_units += [unit_identifier + i for i in supported_special_units]
+    if 'sf' in kwargs.keys():
+        sf = kwargs.get('sf')
 
     # define list of all supported values for the unit isolated list
     supported_units_and_mo = tuple(nsu.list_sorter(supported_units + list(math_ops), reverse=True))
@@ -747,7 +760,7 @@ def unit_parser(math_string, unit_identifier=' ', result='math', **kwargs):
     # if no units are found in the expression, run expression through .parser() and exit
     if not unit_list_SI:
         unit_result = ''
-        math_result = nsp.parser(math_string, steps=False, cprint=parser_print, true_string=parser_true_string)
+        math_result = nsp.parser(math_string, steps=False, cprint=parser_print, true_string=parser_true_string, sf=sf)
         comp_result = str(math_result) + unit_result
         if result == 'math':
             return math_result
@@ -886,7 +899,7 @@ def unit_parser(math_string, unit_identifier=' ', result='math', **kwargs):
         else:
             unit_result = ' ' + abb_res
     math_result = nsp.parser(replaced_unit_string, steps=False, cprint=parser_print, true_string=parser_true_string,
-                             unit_res=unit_result)
+                             unit_res=unit_result, sf=sf)
     comp_result = str(math_result) + unit_result
     if result == 'math':
         return math_result
